t5_generation_config = {
    "do_sample": True,
    "num_beams": 2,
    "repetition_penalty": 5.0,
    "length_penalty": 1.0,
    "num_return_sequences": 1,
}

GPT2_generation_config = {
    "do_sample": True,
    "num_beams": 2,
    "repetition_penalty": 5.0,
    "length_penalty": 1.0,
}